import requests
import uuid

class AlpacaTrading:

    def __init__(self, api_key, api_secret, is_papertrade=True):
        
        self.api_key = api_key
        self.api_secret = api_secret
        self.base_url = 'https://paper-api.alpaca.markets' if is_papertrade else 'https://api.alpaca.markets'

        self.headers = {
            'APCA-API-KEY-ID': self.api_key,
            'APCA-API-SECRET-KEY': self.api_secret
        }

        self.is_papertrade = is_papertrade

    def _send_get_request(self, endpoint, params=None):
        url = self.base_url + endpoint
        response = requests.get(url, headers=self.headers, params=params)
        return response

    def _send_post_request(self, endpoint, data):
        url = self.base_url + endpoint
        response = requests.post(url, json=data, headers=self.headers)
        return response

    def _send_delete_request(self, endpoint, data):
        url = self.base_url + endpoint
        response = requests.delete(url, json=data, headers=self.headers)
        return response

    def get_open_positions(self):
        endpoint = '/v2/positions'
        response = self._send_get_request(endpoint)
        
        if response.status_code == 200:
            return response.json()
        else:
            return None
        
    def get_orders(self, order_id=None):
        endpoint = f'/v2/orders/{order_id}'
        response = self._send_get_request(endpoint)
        
        if response.status_code == 200:
            return response.json()
        else:
            return None
        

    def get_orders_list(self):

        endpoint = '/v2/orders'
        response = self._send_get_request(endpoint)
        
        if response.status_code == 200:
            return response.json()
        else:
            return None
        
    def open_position(self, symbol, quantity, stop_loss, strategy_name, return_client_order_id :bool = False, time_in_force :str = 'gtc'):
        positions = self.get_open_positions()
        
        if positions:
            symbol_filter = symbol.replace('/', '')
            existing_positions = [pos for pos in positions if pos.get('symbol') == symbol_filter]
            if existing_positions:
                print(f'Position for {symbol_filter} already exists')
                return
        
        client_order_id = strategy_name + '@' + str(uuid.uuid4()).replace('-','')[:10]
        endpoint = '/v2/orders'
        order_data = {
            'symbol': symbol,
            'notional': quantity,
            'side': 'buy', 
            'type': 'market',
            'time_in_force': time_in_force,
            'stop_loss': {
                'stop_price': stop_loss
            },
            'client_order_id': client_order_id

        }
        
        response = self._send_post_request(endpoint, order_data)
        
        if response.status_code == 200:
            print(f"Position opened for {symbol}")
            if not return_client_order_id:
                return response.json()
            else:
                return client_order_id
        else:
            return None

    def close_position(self, symbol):
        endpoint = '/v2/positions'
        position_data = {
            'symbol': symbol
        }
        
        response = self._send_delete_request(endpoint, position_data)
        
        if response.status_code == 200:
            return True
        else:
            return False



from ta.trend import SMAIndicator
from pyttrading.utils.logs import log
from scipy.optimize import minimize
import plotly.graph_objects as go
import numpy as np
from plotly.subplots import make_subplots
import os
from pyttrading.utils.pre_processing import remove_noise_trading
from ta.trend import MACD
from pyttrading.backtesting_custom.generic_backtesting import GenericBacktesting


params_default = [(6, 100), (4, 100), (4, 100)]
params_default = [(30, 100), (30, 100), (30, 100)]
params_default = [(100, 200), (100, 200), (100, 200)]



params_init = []
for param in params_default:
    param_low = param[0]
    params_init.append(param_low)


class Strategy: 

    def __init__(self, df=None):
        self.name = 'MACD'
        self.open_long = 1
        self.close_long = 2
        self.keep = 0
    
        
    def plot(self, df=None, show_graph :bool = True, save_figure: bool = False, params=None, title: str = 'Title figure'):
        
        fig = make_subplots(rows=2, cols=1, shared_xaxes=True, vertical_spacing=0.05, row_heights=[0.7, 0.3], subplot_titles=('Price and EMA', 'RSI'))
        fig.add_trace(go.Scatter(x=df.index, y=df['close'], mode='lines', name='Close Price', line=dict(color='blue')), row=1, col=1)
        buy_indices = df.index[df['actions'] == 1]
        sell_indices = df.index[df['actions'] == 2]
        fig.add_trace(go.Scatter(x=buy_indices, y=df['close'].loc[buy_indices], mode='markers', name='Buy', marker=dict(color='green', symbol='circle', size=10)), row=1, col=1)
        fig.add_trace(go.Scatter(x=sell_indices, y=df['close'].loc[sell_indices], mode='markers', name='Sell', marker=dict(color='red', symbol='circle', size=10)), row=1, col=1)
        
        fig.add_trace(go.Scatter(x=df.index, y=df['macd_line'], mode='lines', name='EMA', line=dict(color='red', width=1)), row=2, col=1)
        # fig.add_trace(go.Scatter(x=df.index, y=df['macd_line'], mode='lines', name='EMA_S', line=dict(color='yellow', width=1)), row=2, col=1)

        fig.add_trace(go.Scatter(x=buy_indices, y=df['macd_line'].loc[buy_indices], mode='markers', name='Buy', marker=dict(color='green', symbol='circle', size=10)), row=2, col=1)
        fig.add_trace(go.Scatter(x=sell_indices, y=df['macd_line'].loc[sell_indices], mode='markers', name='Sell', marker=dict(color='red', symbol='circle', size=10)), row=2, col=1)

        # fig.add_trace(go.Scatter(x=df.index, y=[params[0]] * len(df), mode='lines', name='Buy Threshold', line=dict(color='green', width=1, dash='dash')), row=2, col=1)
        # fig.add_trace(go.Scatter(x=df.index, y=[params[1]] * len(df), mode='lines', name='Sell Threshold', line=dict(color='blue', width=1, dash='dash')), row=2, col=1)

        fig.update_layout(template='plotly_dark', title=title, xaxis_title='Date', height=600)

        if show_graph:
            fig.show()

        if save_figure:
            if not os.path.exists('tmp'):
                os.makedirs('tmp')
            filename = f'figure_{self.name}.html'
            fig.write_html(os.path.join('tmp', filename))
            filename = f'figure_{self.name}.png'
            fig.write_image(os.path.join('tmp', filename))

    def eval(self, df=None, params=[10, 30]):
        
        window_slow, window_fast, window_sign = params
        
        # window_slow = int(window_slow)
        # window_fast = int(window_fast)
        # window_sign = int( window_sign)

        window_slow = window_slow
        window_fast = window_fast
        window_sign = window_sign
        

        
        macd = MACD(df['close'], 
                    window_slow=window_slow, 
                    window_fast=window_fast, 
                    window_sign=window_sign,
                    fillna=True
                )
        
        df['macd_line'] = macd.macd()
        df['macd_signal'] = macd.macd_signal()
        
        df.loc[(df['macd_line'] > df['macd_signal']) & (df['macd_line'].shift() <= df['macd_signal'].shift()), 'actions'] = 1  # Señal de compra
        df.loc[(df['macd_line'] < df['macd_signal']) & (df['macd_line'].shift() >= df['macd_signal'].shift()), 'actions'] = 2  # Señal de venta
        
        df['actions'] = df['actions'].mask(df['actions'].eq(df['actions'].shift()))
        df['actions'] = df['actions'].fillna(0)
        df['actions'] = remove_noise_trading(actions=df['actions'])
        
        
        return df
    
    def objective_function(self, params=[10,30], df=None):
        
        # Apply the strategy with the current parameters
        
        result_df = self.eval(df=df.copy(), params=params)
        
        back_testing = GenericBacktesting(
            df=result_df,
            skip=True,
            initial_money=2000.0,
            commission=0.02,
            plot_result=False,
            path_save_result=".",
            print_stacks=False
        )
        return_data, _ = back_testing.calculate_bk()
        
        log.info(f'Strategy: {self.name} RETURN: {return_data} PARAMS: {params}')
        # Minimize the negative return to maximize the return
        return -return_data
    
    def optimize(self, parms: list = params_default, initial_guess: list = params_init, df=None, method: str = "Nelder-Mead"):
        
        threshold_bounds = parms
        
        result = minimize(self.objective_function, initial_guess, bounds=threshold_bounds, args=(df), method=method)
        
        best_breakout_threshold = result
        
        best_return = -result.fun

        return best_return, best_breakout_threshold
    


    def experiment(self, df=None, parms=params_default,initial_guess=params_init):
        
        methods_list = [
            # 'Nelder-Mead',
            'Powell',
            # 'CG',
            # 'L-BFGS-B',
            # 'COBYLA',
            # 'trust-constr'
        ]
        
        results_methods = {}
        results_method_list = []
        result_method_name = []
        
        for method in methods_list:
            print(f"Method: {method}")
    
            best_return, best_sma_threshold = self.optimize(
                parms=parms, 
                initial_guess=initial_guess,
                df=df, 
                method=method
            )
    
            results_method_list.append(best_return)
            result_method_name.append(method)
    
            results_methods[method] = best_return, best_sma_threshold.x
            

        # get the max value of results_method_list
        optimize = results_method_list.index(max(results_method_list))
        method = result_method_name[optimize]
        best_return, best_sma_threshold = results_methods[method]
        
        return method, best_return, best_sma_threshold

# strategy_data = Strategy(df=data)
# method, best_return, best_parameters = strategy_data.experiment(df=data)
# print(f"Best Return: {best_return}", )
# print(f"Best parms: {best_parameters}")
# print(f"Method: {method}")

# data2 = strategy_data.eval(df=data, params=best_parameters)
# data2.head()
# strategy_data.plot(df=data2, save_figure=True)




import pandas as pd
import matplotlib.pyplot as plt
import os


def df_actions_market_value(df=None, capital :float = 2000, save_figure :bool = False, show_figure :bool = False, fig_path = 'tmp/market_value_plot.png', title='Strategy Market' ):

    df['market_value'] = 0
    price_buy = None
    price_buy = None
    qty_buy = None

    for i, action in enumerate(df['actions']):

        if action == 1:  # Compra
            price_buy =  df['close'][i] #TODO verify
            # price_buy = df['close'].iloc[i]

            qty_buy = capital/price_buy
        elif action == 2 and price_buy:  # Venta
            price_buy = None
            qty_buy = None
        
        if qty_buy:
            capital = qty_buy * df['close'][i] #TODO verify
            # capital = qty_buy * df['close'].iloc[i]


        df['market_value'][i] = capital
        # df.at[i, 'market_value'] = capital 

    if os.path.exists(fig_path):
        os.remove(fig_path)

    if save_figure:
        plt.figure(figsize=(10, 6))
        plt.plot(df.index, df['market_value'], color='b')
        plt.title(title)
        plt.xlabel('Date')
        plt.ylabel('Market Value')
        plt.xticks(rotation=45)
        plt.grid(True)
        plt.tight_layout()
        plt.savefig(fig_path)
        if show_figure:
            plt.show()

    return df



import mlflow 

class FakeModel(mlflow.pyfunc.PythonModel):
    def load_context(self, context):
        pass

    def predict(self, context, model_input):
        return self.model.predict(model_input)

class FictitiousModel:
    def predict(self, input_data):
        return input_data * 2


import json
from backtesting import Backtest, Strategy
import os
import json
from backtesting import Backtest, Strategy
from pyttrading.utils.logs import log
import pandas as pd


class DataBacktestingLong(Strategy):

    def init(self):

        self.buy_signal = False
        self.sell_signal = False

    def next(self):
        last_action = self.data.actions[-1] 
        if last_action == 1:
            self.position.close()
            self.buy()
        elif last_action == 2:
            self.position.close()
            self.sell()

class DataBacktestingShort(Strategy):

    def init(self):

        self.buy_signal = False
        self.sell_signal = False

    def next(self):
        last_action = self.data.actions[-1]
        if last_action == 2:
            self.position.close()
            self.sell()
        elif last_action == 1:
            self.position.close()
            self.buy()


class GenericBacktesting:


    def __init__(self, df, skip=True, initial_money=2000.0, commission=0.02, plot_result=True, path_save_result=".", print_stacks :bool=False, bk_type :str = 'long', save_bk_figure :bool = False):

        df.index = pd.to_datetime(df.index)

        self.df = df
        self.skip = skip
        self.initial_money = initial_money
        self.commission = commission
        self.plot_result = plot_result
        self.path_save_result = path_save_result
        self.print_stacks = print_stacks
        self.bk_type = bk_type

        self.save_bk_figure = save_bk_figure

    def rename_columns(self):

        self.df_bk = self.df.rename(columns={
                'open': 'Open',
                'high': 'High',
                'low': 'Low',
                'close': 'Close',
                'volume': 'Volume'
            })

    def calculate_bk(self):

        self.rename_columns()
        self.df_bk_pandas = pd.DataFrame(self.df_bk)
        
        if self.bk_type == 'long':
            strategy_use = DataBacktestingLong
        elif self.bk_type == 'short':
            strategy_use = DataBacktestingShort
        else: 
            raise ValueError('bk_type not valid, need to do long or short')


        bt = Backtest(
                    data=self.df_bk_pandas, 
                    strategy=strategy_use,
                    cash=self.initial_money,
                    commission=self.commission,
                    exclusive_orders=True
                )

        stats = bt.run()
        stats_json = json.loads(stats.to_json())
        if self.print_stacks:
            print(stats)

        if self.save_bk_figure:
            graph_tmp = f'tmp/backtest_{self.bk_type}.html'
            if os.path.exists(graph_tmp):
                os.remove(graph_tmp)

            bt.plot(filename=graph_tmp, open_browser=False)

        profit_factor = stats['Equity Final [$]'] / self.initial_money

        return profit_factor, stats_json
    


def get_backtesting(df=None, initial_money :float = 2000.0, bk_type :str = 'long'):

    back_testing = GenericBacktesting(
                    df=df,
                    skip=True,
                    initial_money=initial_money,
                    commission=0.02,
                    plot_result=False,
                    path_save_result='tmp',
                    print_stacks=False,
                    bk_type=bk_type
                )

    return_data, stats_json = back_testing.calculate_bk()
    
    return return_data, stats_json 





from scipy.optimize import minimize
from ..backtesting_custom.generic_backtesting import get_backtesting
import plotly.graph_objs as go
from scipy.optimize import minimize
from pyttrading.backtesting_custom.generic_backtesting import get_backtesting

def stop_loss_simple(data=None, stop_loss=-0.05, init_capital :float = 2000, column_action_name :str = 'actions', sell_value :int = 2):
    
    price_buy = None
    qty_buy = None
    capital = init_capital

    dff = data.copy()
    for i, action in enumerate(dff['actions']):
        if action == 1:
            # price_buy = dff['close'][i]
            price_buy = dff['close'].iloc[i]
            price_buy = float(price_buy)
            capital = float(capital)

            qty_buy = capital/price_buy

        if qty_buy:
            # current_capital = qty_buy * dff['close'][i]
            current_capital = qty_buy * dff['close'].iloc[i]

            equity = (current_capital-capital)/capital

            if equity < stop_loss:
                dff.loc[dff.index[i], column_action_name] = sell_value

                qty_buy = None
                
                try:
                    filtered_df = dff.iloc[i+1:]
                    idx = filtered_df.index[filtered_df["actions"] == 2][0]
                    dff.at[idx, 'actions'] = 0
                except:
                    pass
        
    return dff


# Optimization Stop Loss

def objective_function(stop_loss, df_original, init_capital=2000, sell_value=2):

    df_stop_loss = stop_loss_simple(data=df_original, stop_loss=stop_loss, init_capital=init_capital, sell_value=sell_value)
    return_data, _ = get_backtesting(df=df_stop_loss)

    print(f"SEARCH THE BEST STOP LOSS SL: {stop_loss} RETURN: {return_data}")
    return -return_data

def optimize_stop_loss_simple(df_original, st_from=0.01, st_to=10):

    #method='Nelder-Mead'
    method='Powell'
    result = minimize(objective_function, x0=0, args=(df_original,), bounds=[(st_from, st_to)], method=method, options={'maxiter': 1000})
    best_stop_loss = result.x[0]
    best_return = -result.fun
    
    return best_stop_loss, best_return




class TrailingStopLoss:

    default_params = [0.05, 0.05]

    def __init__(self, init_capital: float = 2000, column_action_name: str = "actions", sell_value: int = 2, data=None):
        self.init_capital = init_capital
        self.column_action_name = column_action_name
        self.sell_value = sell_value
        self.data = data

    def stop_loss(self, params: list = default_params):
        stop_loss_constant, stop_loss_trailing = params

        qty_buy = None
        capital = self.init_capital
        start_capital = self.init_capital

        dff = self.data.copy()
        for i, action in enumerate(dff["actions"]):
            if i > 1:

                price_current = dff["close"].iloc[i]
                price_last = dff["close"].iloc[i-1]

                if action == 1:
                    qty_buy = capital / price_current

                if qty_buy:
                    capital_current = price_current * qty_buy
                    if price_current  >= price_last and capital_current >= start_capital:
                        capital = capital_current
                        qty_buy = capital / price_current
                    else:
                        equity_constant = (capital_current-start_capital)/start_capital

                        activate_stop_loss = False
                        if equity_constant < stop_loss_constant:
                            activate_stop_loss = True

                        equity = (capital_current-capital)/capital

                        if equity < stop_loss_trailing:
                            activate_stop_loss = True

                        if activate_stop_loss:
                            start_capital = capital
                            dff.loc[dff.index[i], self.column_action_name] = self.sell_value
                            qty_buy = None
                            try:
                                filtered_df = dff.iloc[i + 1:]
                                idx = filtered_df.index[filtered_df["actions"] == 2][0]
                                dff.at[idx, 'actions'] = 0
                            except:
                                pass

        return dff

    def objective_function(self, params: list = default_params):
        df_stop_loss = self.stop_loss(params=params)
        return_data, _ = get_backtesting(df=df_stop_loss)

        return -return_data

    def optimize(self, params: list = default_params):

        params_init = []
        for param in params:
            param_low = param[0]
            params_init.append(param_low)

        method = "Powell"
        result = minimize(self.objective_function,x0=params_init,  bounds=params, method=method, options={"maxiter": 1000})

        best_parameters = result.x
        best_return = -result.fun

        return best_return, best_parameters




class ConstantStopLoss:

    default_params = [0.05]


    def __init__(self, init_capital: float = 2000, column_action_name: str = "actions", sell_value: int = 2, data=None):
        
        self.init_capital = init_capital
        self.column_action_name = column_action_name
        self.sell_value = sell_value
        self.data = data

    def stop_loss(self, params: list = default_params):

        stop_loss = params
        price_buy = None
        qty_buy = None
        capital = self.init_capital

        dff = self.data.copy()
        for i, action in enumerate(dff['actions']):
            if action == 1:
                price_buy = dff['close'].iloc[i]
                price_buy = float(price_buy)
                capital = float(capital)

                qty_buy = capital/price_buy

            if qty_buy:
                current_capital = qty_buy * dff['close'].iloc[i]

                equity = (current_capital-capital)/capital

                # if equity < stop_loss:
                if any(equity < stop_loss):
                    dff.loc[dff.index[i], self.column_action_name] = self.sell_value
                    qty_buy = None
                    
                    try:
                        filtered_df = dff.iloc[i+1:]
                        idx = filtered_df.index[filtered_df["actions"] == 2][0]
                        dff.at[idx, 'actions'] = 0
                    except:
                        pass
            
        return dff

    def objective_function(self, params = default_params):

        df_stop_loss = self.stop_loss( params=params)
        return_data, _ = get_backtesting(df=df_stop_loss)

        print(f"SEARCH THE BEST STOP LOSS SL: {params} RETURN: {return_data}")
        return -return_data

    def optimize(self, params: list = default_params):

        params_init = []
        for param in params:
            param_low = param[0]
            params_init.append(param_low)

        method = "Powell"
        result = minimize(self.objective_function,x0=params_init,  bounds=params, method=method, options={"maxiter": 1000})

        best_parameters = result.x
        best_return = -result.fun

        return best_return, best_parameters



import matplotlib.pyplot as plt
import plotly.graph_objs as go


def generate_market_value_plot(df_action_market, stop_loss_constant, stop_loss_trailing):

    # Data traces
    trace_original = go.Scatter(
        x=df_action_market.index,
        y=df_action_market['market_value'],
        mode='lines',
        name='Original'
    )

    sl_constant = go.Scatter(
        x=stop_loss_constant.index,
        y=stop_loss_constant['market_value'],
        mode='lines',
        name='SL'
    )

    trace_trailing = go.Scatter(
        x=stop_loss_trailing.index,
        y=stop_loss_trailing['market_value'],
        mode='lines',
        name='SL Trailing'
    )

    # Create layout with 'plotly_dark' template
    layout = go.Layout(
        title='Operation Values',
        xaxis=dict(title='Fecha'),
        yaxis=dict(title='Valor de Mercado'),
        template='plotly_dark'
    )

    # Combine data traces and layout, then plot
    fig = go.Figure(data=[trace_original, sl_constant,trace_trailing], layout=layout)

    fig.write_html('tmp/market_value_plot.html')

    # fig.show()

    return 'tmp/market_value_plot.png'

def generate_actions_plot(df_action_market):
    
    actions_1 = df_action_market[df_action_market['actions'] == 1]
    actions_2 = df_action_market[df_action_market['actions'] == 2]
    actions_2_sl = df_action_market[df_action_market['actions'] == 2]

    plt.figure(figsize=(10, 6))
    plt.scatter(actions_1.index, actions_1['open'], color='green', label='BUY')
    plt.scatter(actions_2.index, actions_2['open'], color='red', label='SELL')
    plt.scatter(actions_2_sl.index, actions_2_sl['open'], color='orange', label='StopLoss')
    plt.plot(df_action_market.index, df_action_market['close'], label='Close', linestyle='-')
    plt.xlabel('Index')
    plt.ylabel('Values')
    plt.title('Scatter Plot for Actions 1 and 2 with Closing Line')
    plt.legend()
    plt.savefig('tmp/actions_plot.png')
    plt.close()  # Close the figure to free memory

    return 'tmp/actions_plot.png'


from importlib import import_module
import json
import os
from ..backtesting_custom import GenericBacktesting

class ModelSelector: 
    
    def __init__(self, 
                model_name: str= "macd", 
                model_folder: str="strategies",
                type_model: str="basic",
                configuration=None,
                path_model :str ="",
                symbol :str = "",
                df=None,
                interval=None, 
                mlflow=None,
                bk_initial_money = 2000,
                bk_commission=0.02

            ):
        
        self.mlflow = mlflow
        self.model_name = model_name
        self.model_folder = model_folder
        self.type_model = type_model
        self.configuration = configuration
        self.df = df
        self.path_model = path_model
        self.symbol = symbol
        self.interval = interval

        self.bk_initial_money = bk_initial_money
        self.bk_commission = bk_commission
        
        self.strategy_id=f"{self.model_name}_{symbol.lower()}_{interval}"

    def init_tmp_path(self):
         
        image_folder = self.path_model + "/images"

        if not os.path.exists(image_folder):
            os.mkdir(image_folder)
    
    def run_experiments(self):

        self.basic_strategies  = self.basic_models()
        self.strategy = self.basic_strategies (
                    strategy_name=self.model_name
                )
        method, best_return, best_parameters = self.strategy.experiment(
                    df=self.df
                )

        return method, best_return, best_parameters

    def save_mlflow(self):


        self.mlflow.set_tag("strategy", self.model_name)
        self.mlflow.set_tag("method", self.params.get('method'))
        self.mlflow.log_param("optimized_params", json.dumps(self.params))
        self.mlflow.log_metric('best_return',  self.params.get('best_return'))

        for key, value in self.stats_json.items():
            key_s = key.replace(" ",'').replace('[%]','').replace('.','').replace('&','').replace('[$]','').replace('(','').replace(')','').replace('#','')
            if '_' not in key_s:
                try:
                    self.mlflow.log_metric(key_s, value)
                except Exception:
                    pass

    def get_backtesting(self, bk_type='short', df=None):

        back_testing_short = GenericBacktesting(
                    df=df,
                    skip=True,
                    initial_money=self.bk_initial_money,
                    commission=self.bk_commission,
                    plot_result=False,
                    path_save_result=self.path_model,
                    print_stacks=False,
                    save_bk_figure=True,
                    bk_type=bk_type
                )

        best_return, stats_json = back_testing_short.calculate_bk()

        return best_return, stats_json

    def select(self):
        best_return = None
        best_return_short = None


        method, best_return, best_parameters = self.run_experiments()

        self.params = {
                    "type": self.type_model,
                    "method": method,
                    "best_return": best_return,
                    "best_parameters": list(best_parameters),
                    "params": list(best_parameters), 
                    "strategy": self.model_name
                }


        df_actions = self.strategy.eval(df=self.df, params=best_parameters)

        return_data, stats_json = self.get_backtesting(df=df_actions, bk_type='long')
        self.stats_json = stats_json
        self.save_mlflow()
        self.init_tmp_path()

        back_testing = GenericBacktesting(
                    df=df_actions,
                    skip=True,
                    initial_money=self.bk_initial_money,
                    commission=self.bk_commission,
                    plot_result=False,
                    path_save_result=self.path_model,
                    print_stacks=False,
                    save_bk_figure=True
                )

        return_data, self.stats_json = back_testing.calculate_bk()
        self.strategy.plot(df=df_actions, save_figure=True, show_graph=False, params=best_parameters, title=f"Strategy: {self.model_name} Return: {return_data}")

        best_return_short, _ = self.get_backtesting(df=df_actions, bk_type='short')
                
        return self.basic_strategies, self.params, best_return, best_return_short
        
        
    def basic_models(self):
        
        libray_trading = f"pyttrading.strategies.{self.model_name}"
        model_module = import_module(libray_trading)
        strategy = getattr(model_module, "Strategy")
        
        return strategy
    
    
    def execute(self, df):
        
        _strategy, params, best_return, best_return_short = self.select()
        strategy = _strategy()
        df_action = strategy.eval(df=df, params=params.get('params'))
        
        return best_return, df_action, best_return_short
        
    
    
    
    
     
    
    
    



def strategy_quality(config, data):

    if not config.use_quality_model:
        print(" [ ⚠️ MODEL QUALITY]: Disabled")
        return True

    metrics = data.metrics
    save_mode = False
    trades = metrics.get('Trades')
    profit_factor = metrics.get('ProfitFactor')

    print(f" [ 🟦 MODEL QUALITY]: TRADES: {trades} >= {config.min_trades}")
    print(f" [ 🟦 MODEL QUALITY]: PROFIT_FACTOR: {profit_factor} >= {config.min_profit_factor}")
    print(f" [ 🟦 MODEL QUALITY]: BEST_RETURN: {metrics.get('best_return_sl') } >= {metrics.get('best_return')}")

    if trades >= config.min_trades and profit_factor >= config.min_profit_factor:

        if metrics.get('best_return_sl') > metrics.get('best_return'):
            save_mode = True

    return save_mode


from scipy.optimize import minimize
from pyttrading.backtesting_custom.generic_backtesting import get_backtesting

class ConstantStopLoss:

    default_params = [0.05]


    def __init__(self, init_capital: float = 2000, column_action_name: str = "actions", sell_value: int = 2, data=None):
        
        self.init_capital = init_capital
        self.column_action_name = column_action_name
        self.sell_value = sell_value
        self.data = data

    def stop_loss(self, params: list = default_params):

        stop_loss = params
        price_buy = None
        qty_buy = None
        capital = self.init_capital

        dff = self.data.copy()
        for i, action in enumerate(dff['actions']):
            if action == 1:
                price_buy = dff['close'].iloc[i]
                price_buy = float(price_buy)
                capital = float(capital)

                qty_buy = capital/price_buy

            if qty_buy:
                current_capital = qty_buy * dff['close'].iloc[i]

                equity = (current_capital-capital)/capital

                # if equity < stop_loss:
                if any(equity < stop_loss):
                    dff.loc[dff.index[i], self.column_action_name] = self.sell_value
                    qty_buy = None
                    
                    try:
                        filtered_df = dff.iloc[i+1:]
                        idx = filtered_df.index[filtered_df["actions"] == 2][0]
                        dff.at[idx, 'actions'] = 0
                    except:
                        pass
            
        return dff

    def objective_function(self, params = default_params):

        df_stop_loss = self.stop_loss( params=params)
        return_data, _ = get_backtesting(df=df_stop_loss)

        print(f"SEARCH THE BEST STOP LOSS SL: {params} RETURN: {return_data}")
        return -return_data

    def optimize(self, params: list = default_params):

        params_init = []
        for param in params:
            param_low = param[0]
            params_init.append(param_low)

        method = "Powell"
        result = minimize(self.objective_function,x0=params_init,  bounds=params, method=method, options={"maxiter": 1000})

        best_parameters = result.x
        best_return = -result.fun

        return best_return, best_parameters



# sl = ConstantStopLoss(data=data2,init_capital=2000)
# # best_return, best_parameters_sl = sl.optimize(params=[(-0.1, -0.01)])

# best_parameters_sl = [-0.05]
# stop_loss_df =  sl.stop_loss(params=best_parameters_sl)
# stop_loss_simple = df_actions_market_value(df=stop_loss_df, capital=2000, save_figure=False)
# market = df_actions_market_value(df=data2, capital=2000, save_figure=False)


# import plotly.graph_objs as go

# # Data traces
# trace_original = go.Scatter(
#     x=market.index,
#     y=market['market_value'],
#     mode='lines',
#     name='Original'
# )

# trace_stop_loss = go.Scatter(
#     x=stop_loss_simple.index,
#     y=stop_loss_simple['market_value'],
#     mode='lines',
#     name='SL'
# )

# # Create layout with 'plotly_dark' template
# layout = go.Layout(
#     title='Comparación del Valor de Mercado con y sin Stop Loss',
#     xaxis=dict(title='Fecha'),
#     yaxis=dict(title='Valor de Mercado'),
#     template='plotly_dark'
# )

# # Combine data traces and layout, then plot
# fig = go.Figure(data=[trace_original, trace_stop_loss], layout=layout)
# fig.show()



import numpy as np


def _add_days_ago2(bar_data, days):
    # Asegurarse de que hay suficientes datos para el desplazamiento
    if len(bar_data.close) < days:
        raise ValueError("No hay suficientes datos para el desplazamiento solicitado")

    # Crear el atributo para 'days' días atrás
    shifted_close = np.roll(bar_data.close, days)
    # Establecer los primeros 'days' valores como NaN
    shifted_close[:days] = np.nan

    return shifted_close


def _add_days_ago(data, days):
    for i in range(1, days + 1):
        data[f'd{i}'] = data['close'].shift(i)
    data.dropna(inplace=True)
    data['date'] = data.index
    return data

def days_ago(data, days):
    days = _add_days_ago2(data, days)
    return days




from ...backtesting_custom.generic_backtesting import GenericBacktesting
from ...utils.pre_processing import remove_noise_trading
from scipy.optimize import minimize
from ta.trend import SMAIndicator
from pyttrading.utils.logs import log
from scipy.optimize import minimize
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import os

params_default = [(5, 30), (10, 100)]

params_init = []
for param in params_default:
    param_low = param[0]
    params_init.append(param_low)

class Strategy: 

    def __init__(self, df=None, strategy_name :str = "StrategyName", bk_initial_money :float = 2000.0, bk_commission :float = 0.02):
        self.name = 'SMA'
        self.open_long = 1
        self.close_long = 2
        self.keep = 0
        self.strategy_name = strategy_name
        self.bk_initial_money = bk_initial_money
        self.bk_commission = bk_commission
    
    def plot(self, df=None, show_graph :bool = True, save_figure: bool = False, title :str = 'Title', params=None):
        
        fig = make_subplots(rows=1, cols=1, shared_xaxes=True, vertical_spacing=0.05, subplot_titles=('Price and EMA', 'RSI'))
        fig.add_trace(go.Scatter(x=df.index, y=df['close'], mode='lines', name='Close Price', line=dict(color='blue')), row=1, col=1)
        buy_indices = df.index[df['actions'] == 1]
        sell_indices = df.index[df['actions'] == 2]
        fig.add_trace(go.Scatter(x=buy_indices, y=df['close'].loc[buy_indices], mode='markers', name='Buy', marker=dict(color='green', symbol='circle', size=10)), row=1, col=1)
        fig.add_trace(go.Scatter(x=sell_indices, y=df['close'].loc[sell_indices], mode='markers', name='Sell', marker=dict(color='red', symbol='circle', size=10)), row=1, col=1)
        fig.add_trace(go.Scatter(x=df.index, y=df['fast_ma'], mode='lines', name='fast_ma', line=dict(color='red', width=1)), row=1, col=1)
        fig.add_trace(go.Scatter(x=df.index, y=df['slow_ma'], mode='lines', name='slow_ma', line=dict(color='yellow', width=1)), row=1, col=1)

        fig.update_layout(template='plotly_dark', title=title, xaxis_title='Date', height=600)
        if show_graph:
            fig.show()

        if save_figure:
            if not os.path.exists('tmp'):
                os.makedirs('tmp')
            filename = f'figure_{self.name}.html'
            fig.write_html(os.path.join('tmp', filename))
            filename = f'figure_{self.name}.png'
            fig.write_image(os.path.join('tmp', filename))

    def eval(self, df=None, params=[10, 30]):
        
        fast_ma_period, slow_ma_period = params
        
        df['fast_ma'] = SMAIndicator(df['close'], int(fast_ma_period), True).sma_indicator()
        df['slow_ma']  = SMAIndicator(df['close'], int(slow_ma_period), True).sma_indicator()
        df['actions'] = self.keep
        df.loc[df['fast_ma'] > df['slow_ma'], 'actions'] = self.open_long
        df.loc[df['fast_ma'] < df['slow_ma'], 'actions'] = self.close_long
        df['actions'] = df['actions'].mask(df['actions'].eq(df['actions'].shift()))
        df['actions'] = df['actions'].fillna(0)
        df['actions'] = remove_noise_trading(actions=df['actions'])
        
        return df
    
    def objective_function(self, params=params_init, df=None):
        
        # Apply the strategy with the current parameters
        
        result_df = self.eval(df=df.copy(), params=params)
        
        back_testing = GenericBacktesting(
            df=result_df,
            skip=True,
            initial_money=self.bk_initial_money, 
            commission=self.bk_commission,
            plot_result=False,
            path_save_result=".",
            print_stacks=False
        )
        return_data, _ = back_testing.calculate_bk()
        
        log.info(f'Strategy: {self.name} RETURN: {return_data} PARAMS: {params}')
        # Minimize the negative return to maximize the return
        return -return_data
    
    def optimize(self, parms: list =params_default, initial_guess: list = [10, 20], df=None, method: str = "Nelder-Mead"):
        
        threshold_bounds = parms
        
        result = minimize(self.objective_function, initial_guess, bounds=threshold_bounds, args=(df), method=method)
        
        best_breakout_threshold = result
        
        best_return = -result.fun

        return best_return, best_breakout_threshold
    


    def experiment(self, df=None, parms=params_default,initial_guess=params_init):
        
        log.info(f"Start Experiment, Strategy: {self.name} Name: {self.strategy_name}")
        methods_list = [
            # 'Nelder-Mead',
            'Powell',
            # 'CG',
            # 'L-BFGS-B',
            # 'COBYLA',
            # 'trust-constr',
            # 'BFGS',
            # 'TNC',
            # 'SLSQP',
            # 'dogleg',
            # 'trust-ncg',
            # 'trust-exact',
            # 'trust-krylov'
        ]
        
        results_methods = {}
        results_method_list = []
        result_method_name = []
        
        for method in methods_list:
            print(f"Method: {method}")
    
            best_return, best_sma_threshold = self.optimize(
                parms=parms, 
                initial_guess=initial_guess,
                df=df, 
                method=method
            )
    
            results_method_list.append(best_return)
            result_method_name.append(method)
    
            results_methods[method] = best_return, best_sma_threshold.x
            

        # get the max value of results_method_list
        optimize = results_method_list.index(max(results_method_list))
        method = result_method_name[optimize]
        best_return, best_sma_threshold = results_methods[method]
        
        return method, best_return, best_sma_threshold


import numpy as np



def inflection_points(df, threshold=0.5, column_name='rsi_s', result_column :str = 'actions'):
    
    df_inf = df.copy()

    df_inf['date'] = df_inf.index

    df_inf.reset_index(drop=True, inplace=True)

    differences = np.diff(df_inf[column_name])
    inflection_points = np.where(abs(differences) > threshold)[0]
    df_inf['inf_point'] = 0
    df_inf.loc[inflection_points, 'inf_point'] = 1 

    df_inf[result_column] = 0
    
    for point in inflection_points:
        window = 10
        subset = df_inf[column_name].iloc[point:point + window]
        if len(subset) > 1:
            if subset.iloc[0] < subset.iloc[-1]:
                df_inf.at[point, result_column] = 1  # Increasing
            elif subset.iloc[0] > subset.iloc[-1]:
                df_inf.at[point, result_column] = 2  # Decreasing
            else:
                df_inf.at[point, result_column] = 0
        else:
            df_inf.at[point, result_column] = 0


    df.reset_index(drop=True, inplace=True)  # Si no tienen el mismo índice
    df_inf.reset_index(drop=True, inplace=True)
    df['actions'] = df_inf['actions']

    df.index = df_inf['date']
    
    return df


from .stop_loss import *
from .backtesting_custom import *
from .utils import *
from .labels import *
from .indicators import *
from .strategies import *
from .brokers import *
from .market_data import *

import os

if not os.path.exists('tmp'):
    os.makedirs('tmp')



import numpy as np
from numba import njit


def cmma(bar_data, lookback):

    @njit  # Enable Numba JIT.
    def vec_cmma(values):
        # Initialize the result array.
        n = len(values)
        out = np.array([np.nan for _ in range(n)])

        # For all bars starting at lookback:
        for i in range(lookback, n):
            # Calculate the moving average for the lookback.
            ma = 0
            for j in range(i - lookback, i):
                ma += values[j]
            ma /= lookback
            # Subtract the moving average from value.
            out[i] = values[i] - ma
        return out

    # Calculate for close prices.
    cmma = vec_cmma(bar_data.close)
    return cmma



import numpy as np
from pyttrading.utils.logs import log
from scipy.optimize import minimize
import plotly.graph_objects as go
import numpy as np
from plotly.subplots import make_subplots
import os
from pyttrading.utils.pre_processing import remove_noise_trading
import pandas as pd
from ...backtesting_custom.generic_backtesting import GenericBacktesting

params_default = [(0.2, 0.4)]
params_init = 0.3


class Strategy: 

    def __init__(self, df=None, strategy_name :str = "StrategyName"):
        self.name = 'DUMMY'
        self.open_long = 1
        self.close_long = 2
        self.keep = 0
        self.strategy_name = strategy_name
    
    def plot(self, df=None, show_graph :bool = True, save_figure: bool = False, params=None, title: str = 'Title figure'):
        
        fig = make_subplots(rows=1, cols=1, shared_xaxes=True, vertical_spacing=0.05, subplot_titles=('Price and EMA', 'RSI'))
        fig.add_trace(go.Scatter(x=df.index, y=df['close'], mode='lines', name='Close Price', line=dict(color='blue')), row=1, col=1)
        buy_indices = df.index[df['actions'] == 1]
        sell_indices = df.index[df['actions'] == 2]
        fig.add_trace(go.Scatter(x=buy_indices, y=df['close'].loc[buy_indices], mode='markers', name='Buy', marker=dict(color='green', symbol='circle', size=10)), row=1, col=1)
        fig.add_trace(go.Scatter(x=sell_indices, y=df['close'].loc[sell_indices], mode='markers', name='Sell', marker=dict(color='red', symbol='circle', size=10)), row=1, col=1)
        fig.update_layout(template='plotly_dark', title=title, xaxis_title='Date', height=600)

        if show_graph:
            fig.show()

        if save_figure:
            if not os.path.exists('tmp'):
                os.makedirs('tmp')
            filename = f'figure_{self.name}.html'
            fig.write_html(os.path.join('tmp', filename))
            filename = f'figure_{self.name}.png'
            fig.write_image(os.path.join('tmp', filename))

    def eval(self, df=None, params=params_default):

        random_1  = params
        if isinstance(random_1, list):
            random_2 = 0.4 - random_1[0]
        else:
            random_2 = 0.4 - random_1

        random_1 = 0.2
        random_2 = 0.2
        df2 = df.copy()
        probabilities = [0.6, 0.2, 0.2]

        df['actions'] = np.random.choice([0,1,2], size=len(df2), p=probabilities)
        df['actions'] = remove_noise_trading(actions=df['actions'])

        return df
    
    def objective_function(self, params=params_default, df=None):
        
        result_df = self.eval(df=df.copy(), params=params)
        
        back_testing = GenericBacktesting(
            df=result_df,
            skip=True,
            initial_money=2000.0,
            commission=0.02,
            plot_result=False,
            path_save_result=".",
            print_stacks=False
        )
        return_data, _ = back_testing.calculate_bk()
        
        log.info(f'Strategy: {self.name} RETURN: {return_data} PARAMS: {params}')
        # Minimize the negative return to maximize the return
        return -return_data
    
    def optimize(self, params: list = params_default, initial_guess: list = params_init, df=None, method: str = "Nelder-Mead"):
        threshold_bounds = params
        result = minimize(self.objective_function, initial_guess, bounds=threshold_bounds, args=(df,), method=method)
        best_parameters = result.x
        best_return = -result.fun
        return best_return, best_parameters
    
    def experiment(self, df=None, params=params_default, initial_guess=params_init):
        methods_list = [
            # 'Nelder-Mead',
            'Powell',
            # 'CG',
            # 'L-BFGS-B',
            # 'COBYLA',
            # 'trust-constr'
        ]
        results_methods = {}
        results_method_list = []
        result_method_name = []
        
        for method in methods_list:
            print(f"Method: {method}")
            best_return, best_parameters = self.optimize(params=params, initial_guess=initial_guess, df=df, method=method)
            results_method_list.append(best_return)
            result_method_name.append(method)
            results_methods[method] = best_return, best_parameters
        
        optimize = results_method_list.index(max(results_method_list))
        method = result_method_name[optimize]
        best_return, best_parameters = results_methods[method]
        
        return method, best_return, best_parameters


from scipy.optimize import minimize
from pyttrading.backtesting_custom.generic_backtesting import get_backtesting


class TrailingStopLoss:

    default_params = [0.05, 0.05]


    def __init__(self, init_capital: float = 2000, column_action_name: str = "actions", sell_value: int = 2, data=None):
        self.init_capital = init_capital
        self.column_action_name = column_action_name
        self.sell_value = sell_value
        self.data = data

    def stop_loss(self, params: list = default_params):
        trailing_stop_price, trailing_stop_pct = params

        price_buy = None
        qty_buy = None
        capital = self.init_capital
        trailing_stop_price = None

        dff = self.data.copy()
        for i, action in enumerate(dff["actions"]):
            if action == 1:
                price_buy = dff["close"].iloc[i]
                price_buy = float(price_buy)
                capital = float(capital)
                trailing_stop_price = price_buy

                qty_buy = capital / price_buy

            if qty_buy and trailing_stop_price:
                current_capital = qty_buy * dff["close"].iloc[i]
                current_trailing_stop_price = price_buy * (1 - trailing_stop_pct)

                if dff["close"].iloc[i] < current_trailing_stop_price:
                    dff.loc[dff.index[i], self.column_action_name] = self.sell_value

                    qty_buy = None
                    trailing_stop_price = None

                    try:
                        filtered_df = dff.iloc[i + 1:]
                        idx = filtered_df.index[filtered_df["actions"] == 2][0]
                        dff.at[idx, 'actions'] = 0
                    except:
                        pass

            if trailing_stop_price and current_trailing_stop_price > trailing_stop_price:
                trailing_stop_price = current_trailing_stop_price

        return dff

    def objective_function(self, params: list = default_params):
        df_stop_loss = self.stop_loss(params=params)
        return_data, _ = get_backtesting(df=df_stop_loss)

        print(f"SEARCH THE BEST STOP LOSS SL TRAILING: {params} RETURN: {return_data}")
        return -return_data

    def optimize(self, params: list = default_params):

        params_init = []
        for param in params:
            param_low = param[0]
            params_init.append(param_low)

        method = "Powell"
        result = minimize(self.objective_function,x0=params_init,  bounds=params, method=method, options={"maxiter": 1000})

        best_parameters = result.x
        best_return = -result.fun

        return best_return, best_parameters




import schedule
import time
import mlflow
import mlflow.pyfunc
from datetime import datetime, timedelta
import pytz
import os

# 'America/New_York'
def get_datetime_now_trade(timezone :str = 'Europe/Madrid', delay_minutes = 15, days_ago = 100):
    
    ny_timezone = pytz.timezone(timezone)
    now_ny = datetime.now(ny_timezone)
    start_date_ny = now_ny - timedelta(days=days_ago)
    start_date_ny = start_date_ny.replace(hour=0, minute=0, second=0, microsecond=0)
    start_date_ny = start_date_ny.strftime('%m/%d/%Y')
    end_date_ny = now_ny - timedelta(minutes=delay_minutes)
    end_date_ny = end_date_ny.strftime('%m/%d/%Y %H:%M')
 
    return start_date_ny, end_date_ny, now_ny


import yaml
import os
from stockstats import StockDataFrame
from ..market_data.opensearch_collector import OpensearchCollector
import mlflow
from config import Config

def get_configuration(func):
    def wrapper(*args, **kwargs):
        with open('config.yaml', 'r') as file:
            cfg = yaml.safe_load(file)
        return func(cfg, *args, **kwargs)
    return wrapper

def label_read_args(func):
    def wrapper(*args, **kwargs):
        config = Config()
        return func(config, *args, **kwargs)

    return wrapper

def label_declare_environment_config_mlflow(func):
    def wrapper(config, *args, **kwargs):
        uri = config.mlflow_host
        os.environ['MLFLOW_TRACKING_USERNAME'] = config.mlflow_username
        os.environ['MLFLOW_TRACKING_PASSWORD'] = config.mlflow_token
        os.environ['MLFLOW_TRACKING_URI'] = uri
        mlflow.set_tracking_uri(uri)
        return func(config, *args, **kwargs)
    return wrapper

def declare_environment_variables_mlflow(func):
    def wrapper(cfg, *args, **kwargs):
        uri = cfg.get('mlflow').get('host')
        os.environ['MLFLOW_TRACKING_USERNAME'] = cfg.get('mlflow').get('username')
        os.environ['MLFLOW_TRACKING_PASSWORD'] = cfg.get('mlflow').get('token')
        os.environ['MLFLOW_TRACKING_URI'] = uri
        mlflow.set_tracking_uri(uri)
        return func(cfg, *args, **kwargs)
    return wrapper

# Decorador para obtener datos del mercado
def get_market_data(func):
    def wrapper(config, *args, **kwargs):

        symbols_periods = kwargs.get('symbols_periods')

        collector = OpensearchCollector(
            url_opensearch=config.url_opensearch,
            url_playground=config.url_playground,
        )
        return func(config, collector, *args, **kwargs)
    return wrapper

def get_indicators(df=None, indicators=None):

    data = StockDataFrame.retype(df)
    for indicator in indicators:
        data[indicator]

    return data

def label_collect_market_data(func):

    def wrapper(config, *args, **kwargs):
        configuration = kwargs.get("strategy_config")

        collector = OpensearchCollector(
            url_opensearch=config.url_opensearch,
            url_playground=config.url_playground,
            time_amount=configuration.get("time_amount", 10)
        )

        data =  collector.market_data(
                symbols=[configuration.get("symbol")],
                start_date=configuration.get("start_date"),
                end_date=configuration.get("end_date"),
                interval=configuration.get("interval"), 
                is_crypto=configuration.get("is_crypto")
            )

        data = get_indicators(df=data, indicators=configuration.get("indicators"))
        
        return func(data, *args, **kwargs)
    return wrapper


def read_mlflow_model(func):
    def wrapper( *args, **kwargs):
        symbols_periods = kwargs.get('symbols_periods')
        stage = kwargs.get('stage')

        model_uri = f"models:/{symbols_periods}/{stage}"
        model = mlflow.pyfunc.load_model(model_uri)
        
        return func(model,*args, **kwargs)
    return wrapper


def label_get_mlflow_experiment_id(func):
    def wrapper(data, *args, **kwargs):
        strategy_config = kwargs.get("strategy_config")
        experiment_name = f"{strategy_config.get('symbol')}_{strategy_config.get('interval')}_stats"
        
        experiment = mlflow.get_experiment_by_name(experiment_name)

        if experiment is None or experiment.lifecycle_stage == 'deleted':
            if experiment is None:
                experiment_id = mlflow.create_experiment(name=experiment_name)
            else:
                mlflow.tracking.MlflowClient().restore_experiment(experiment.experiment_id)
                experiment_id = experiment.experiment_id
        else:
            experiment_id = experiment.experiment_id

        return func(experiment_id, data, *args, **kwargs)
    return wrapper


def get_mlflow_strategies(func):

    def wrapper(config, *args, **kwargs):

        interval = kwargs.get('interval')
        stage = kwargs.get('stage')

        client = mlflow.tracking.MlflowClient()
        all_strategies = client.search_model_versions()

        strategies = []
        for st in all_strategies:
            if st.current_stage == stage:
                interval_st = st.tags.get('interval')
                stage_st = st.current_stage
            
                if stage_st == stage and interval_st == interval:
                    strategies.append(st)


        return func(strategies, *args, **kwargs)
    return wrapper



from .df_actions_market_value import *
from .plots import *
from .strategy_quality import *
from .fake_model import *
from .pre_processing import *
from .logs import log
from .inflection_points import *
from .get_datetime_now_trade import *


import requests
from requests.auth import HTTPBasicAuth
import json
from datetime import datetime
import logging as log


def register_event(date, data, opensearch_host):

    index_name = IndexList().tt_scheduled
    ops = Opensearch(index_name=index_name, host=opensearch_host)

    data['date'] = date.strftime("%Y-%m-%dT%H:%M:%S+00:00")
    ops.send(
        data=data
    )


class IndexList:
    def __init__(self):
        self.tt_trading_models = 'tt-trading-models'
        self.tt_trades = 'tt-trades'
        self.tt_ml_models = 'tt-ml-models'
        self.tt_scheduled = 'tt-scheduled'
        self.tt_positions = 'tt-positions'

class Opensearch:

    def __init__(self,
                 index_name='tt-trading-models',
                 host :str = None, 
                 username :str = None, 
                 password :str = None
                ):

        if username:
            self.auth = HTTPBasicAuth(self.username, self.password)
        else:
            self.auth = None

        self.host = host
        self.default_header = {"Content-Type": "application/json"}
        self.index_name = index_name

    def _request(self, method :str ='GET',url :str ='/docs'):
        
        response = requests.request(
            method,
            f"{self.host}{self.index_name}{url}",
            headers=self.default_header,
            auth=self.auth,
        )

        return response
    
    def get_one(self, pipeline_id):
        
        response = requests.request(
            "GET",
            f"{self.host}{self.index_name}/_doc/{pipeline_id}",
            headers=self.default_header,
            auth=self.auth,
        )
        response = response.json()
        response = response.get('_source')
        return response
            
    def read(self, query=None):
        """
        Update the status field of a document based on its custom 'id' field.

        Parameters:
            custom_id (str): The custom 'id' of the document.
            new_status (str): The new status to set.
        """
        # Step 1: Search for the document by its custom 'id' to get the OpenSearch _id

        response = requests.request(
            "GET",
            f"{self.host}{self.index_name}/_search",
            headers=self.default_header,
            auth=self.auth,
            json=query
        )

        response.raise_for_status()
        search_results = response.json()
        
        return search_results

    def get_by_id(self, id: str):
        log.info("Opensearch - get_by_id()")

        response = requests.request(
            "GET",
            f"{self.host}{self.index_name}/_doc/{id}",
            headers=self.default_header,
            auth=self.auth,
        )
        response = response.json()

        if response.get('found'):
            response = response.get('_source')
            return response
        else:
            log.warn(f"Record with ID: {id} not found")
            return None

    def send(self, data=None, test_id: str = None):
        
        try:
            _id = data.get('_id')
            if not _id:
                response = requests.request(
                    "POST",  # Changed from PUT to POST for auto-generating IDs
                    self.host + self.index_name + "/_doc/",
                    headers=self.default_header,
                    auth=self.auth,
                    json=data
                )

                log.info("Data sent successfully.")
            else:
                del data['_id']
                log.info(f"User Custom _id {_id}")
                response = requests.request(
                    "PUT",  # Changed from PUT to POST for auto-generating IDs
                    self.host + self.index_name + f"/_doc/{_id}",
                    headers=self.default_header,
                    auth=self.auth,
                    json=data
                )
            
            response.raise_for_status()
            return response
        except Exception as e:
            log.error(data)
            log.error(f"Failed to send data: {e}")

    def create_index(self, index_name: str = "my-first-request", schema=None):

        try:
            response = requests.request(
                "PUT",
                self.host + index_name,
                headers=self.default_header,
                auth=self.auth,
                data=json.dumps(schema)
            )
            
            
            if response.status_code == 400:
                log.warn('Init Index -------')
                log.warn(str(response.text))
                log.warn('-------------------')
                return
            response.raise_for_status()
            log.info(f"Index {index_name} created successfully.")
        except Exception as e:
            log.warn(f"Failed to create index: {e}")

    def get_index_records(self):
        try:
            response = requests.request(
                "GET",
                f"{self.host}{self.index_name}/_search",
                headers=self.default_header,
                auth=self.auth
            )

            response.raise_for_status()
            response = response.json()
            return response
        except Exception as e:
            log.error(f"Query failed: {e}")
            return

    def query_index(self, index_name: str=None, query=None):

        if index_name is None:
            index_name = self.index_name

        if query is None:
            query = {
                "query": {
                    "match_all": {}
                }
            }

        try:
            response = requests.request(
                "GET",
                f"{self.host}{index_name}/_search",
                headers=self.default_header,
                auth=self.auth,
                json=query
            )

            response.raise_for_status()
            log.info(response.json())
            return response.json()
        except Exception as e:
            log.error(f"Query failed: {e}")
            return

    def prepare_bulk_data(self, documents):
        bulk_data = ""
        for doc in documents:
            action = {"index": {"_index": self.index_name}}
            bulk_data += json.dumps(action) + "\n"
            bulk_data += json.dumps(doc["data"]) + "\n"
        return bulk_data

    def update_bulk(self, documents):
        
        try:
            url = f"{self.host}_bulk"
            log.info(url)
            bulk_data = self.prepare_bulk_data(documents)

            response = requests.request(
                "POST",
                url,
                headers=self.default_header,
                auth=self.auth,
                data=bulk_data
            )
            response.raise_for_status()

        except requests.RequestException as e:
    
            log.error(f"Error to send bulk: {e}")


    def update(self, document_id, data, date=None):

        try:
            
            url = f"{self.host}{self.index_name}/_doc/{document_id}"
            
            if not date:
                try:

                    now = datetime.utcnow()
                    now = now.strftime("%Y-%m-%dT%H:%M:%S+00:00")
                    log.info(f"[OPENSEARCH DATE FORMAT 1] {now}")
                    
                    data['date'] = now     
                    response = requests.request(
                        "PUT",
                        url,
                        headers=self.default_header,
                        auth=self.auth,
                        json=data
                    )
                    response.raise_for_status()
                except Exception as e:
                    now = datetime.now()
                    now = int(now.timestamp() * 1000)
                    log.info(f"[OPENSEARCH DATE FORMAT 1] {now}")
                    data['date'] = now
                    response = requests.request(
                        "PUT",
                        url,
                        headers=self.default_header,
                        auth=self.auth,
                        json=data
                    )

            else:
                data['date'] = date
                now = date
                log.info(f"[OPENSEARCH DATE FORMAT 3] {now}")
                response = requests.request(
                        "PUT",
                        url,
                        headers=self.default_header,
                        auth=self.auth,
                        json=data
                    )
            
            
            datetime_now = now
            log.info(f'[OPENSEARCH] Save one opensearch: {url} date: {datetime_now}')
            log.info(f'[OPENSEARCH] data: {data}')
            
            response = requests.request(
                "PUT",
                url,
                headers=self.default_header,
                auth=self.auth,
                json=data
            )
            
            if response.status_code == 400:
                log.warn("Not was saved on opensearch the data")
                log.warn(response.text)
                
            response.raise_for_status()
            log.info("Success Update opensearch document.")
            return response
        except Exception as e:
            log.error(f"Opensearch update failed: {e}")
            raise RuntimeError(f"NOT WAS POSSIBLE SAVE THE DATA ON OPENSEARCH, ADD THE FORMAT OF DATE ON OPENSEARCH INDEX {self.index_name} {str(e)}")

    def delete(self, document_id):

        try:

            query = {
                "query": {
                    "match": {
                        "_id": document_id
                    }
                }
            }

            response = requests.request(
                "POST",
                f"{self.host}{self.index_name}/_delete_by_query",
                headers=self.default_header,
                auth=self.auth,
                json=query
            )

            response.raise_for_status()
            log.info(response.json())

        except Exception as e:
            log.error(f"Opensearch update failed: {e}")
            return None

    def delete_all_data(self, index_name=None):

        query = {
            "query": {
                "match_all": {}
            }
        }
        
        response = requests.request(
                "POST",
                f"{self.host}{index_name}/_delete_by_query",
                headers=self.default_header,
                auth=self.auth,
                json=query
            )
        return response



class TradingModels:

    def schema(self):
        return {
            "settings": {
                "number_of_shards": 1,
                "number_of_replicas": 0
            },
            "mappings": {
                "properties": {
                    "date": {
                        "type": "date"
                        },
                    "name": {"type": "keyword"},
                    "status": {"type": "keyword"},
                    "best_return": {"type": "float"},
                    "best_return_sl": {"type": "float"},
                    "trades": {"type": "integer"},
                    "symbol": {"type": "keyword"},
                    "strategy": {"type": "keyword"},
                    "interval": {"type": "keyword"},
                    "is_crypto": {"type": "bool"},
                }
            }
        }


class MLModels:

    def schema(self):
        return {
            "settings": {
                "number_of_shards": 1,
                "number_of_replicas": 0
            },
            "mappings": {
                "properties": {
                    "date": {
                        "type": "date"
                        },
                    "name": {"type": "keyword"},
                    "stage": {"type": "keyword"},
                    "status": {"type": "keyword"},
                    "best_return": {"type": "float"},
                    "best_return_sl": {"type": "float"}
                }
            }
        }


class TTScheduled:

    def schema(self):
        return {
            "settings": {
                "number_of_shards": 1,
                "number_of_replicas": 0
            },
            "mappings": {
                "properties": {
                    "date": {
                        "type": "date"
                        },
                    "name": {"type": "keyword"},
                    "close": {"type": "float"},
                    "interval": {"type": "keyword"},
                    "action": {"type": "keyword"},
                    "client_order_id":  {"type": "keyword"}
                }
            }
        }



class TTPositions: 

    def schema(self):
        return {
            "settings": {
                "number_of_shards": 1,
                "number_of_replicas": 0
            },
            "mappings": {
                "properties": {
                    "date": {
                        "type": "date"
                        },
                    "exchange": {"type": "keyword"},
                    "symbol": {"type": "keyword"},
                    "qty": {"type": "float"},
                    "market_value": {"type": "float"},
                    "cost_basis": {"type": "float"},
                    "asset_id": {"type": "text"},
                    "profit": {"type": "float"}
                }
            }
        }


# REGISTER INDEX ON OPENSEARCH
from config import Config

settings = Config()
index_name = IndexList().tt_positions
schema = TTPositions().schema()
ops_tt_positions = Opensearch(index_name=index_name, host=settings.opensearch_host)
ops_tt_positions.create_index(index_name=index_name, schema=schema)

index_name = IndexList().tt_scheduled
schema = TTScheduled().schema()
ops_tt_scheduled = Opensearch(index_name=index_name, host=settings.opensearch_host)
ops_tt_scheduled.create_index(index_name=index_name, schema=schema)

index_name_ml_models = IndexList().tt_ml_models
schema = MLModels().schema()
ops_tt_models = Opensearch(index_name=index_name_ml_models, host=settings.opensearch_host)
ops_tt_models.create_index(index_name=index_name_ml_models, schema=schema)